/* Software-based Trusted Platform Module (TPM) Emulator
 * Emulator Copyright (C) 2004-2010 Mario Strasser <mast@gmx.net>
 * Timing/plugin code Copyright (C) 2010-2012 Stephen R. Tate <srtate@uncg.edu>
 *
 * This module is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 *
 * This module is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 */

#include "tpm/tpm_marshalling.h"
#include "tpm/plugins.h"

/*                               */
/*    Null (dummy) functions     */
/*                               */


/**********************************************************************
 * .._is_genkey() means this keytype/algorithm can be used for Sign, Seal, 
 * Unbind, etc. operations
 */
static BOOL null_is_genkey(void)
{
  return FALSE;
}


/**********************************************************************
 * .._is_transkey() means this keytype/algorithm can be used for
 * EstablishTransport, ExecuteTransport, and ReleaseTransportSigned
 */
static BOOL null_is_transkey(void)
{
  return FALSE;
}


/**********************************************************************
 * .._es_is_validtrans returns true is this is a valid encryption scheme (ES)
 * for a transport session.  Actually only used when getting capability
 * TPM_CAP_TRANS_ES
 */
static BOOL null_es_is_validtrans(TPM_ENC_SCHEME es)
{
  return FALSE;
}


/**********************************************************************
 * .._parms_is_supported returns true if the parameters in the parms
 * structure supported?  (key size, etc.)
 */
static BOOL null_parms_is_supported(TPM_KEY_PARMS *parms)
{
  return FALSE;
}


/**********************************************************************
 * .._parms_is_minimal_size returns true if the keysize is at least some
 * minimal reasonable value
 */
static BOOL null_parms_is_minimal_size(TPM_KEY_PARMS *parms)
{
  return FALSE;
}


/**********************************************************************
 * .._parms_is_fips_size returns true if the keysize is at least some
 * the size required by FIPS requirements (maybe be more than the minimal
 * size otherwise supported).
 */
static BOOL null_parms_is_fips_size(TPM_KEY_PARMS *parms)
{
  return FALSE;
}


/**********************************************************************
 * .._parms_is_strong_size returns true if the keysize is sufficient to
 * support strong security (e.g., 2048 bits for an RSA key)
 */
static BOOL null_parms_is_strong_size(TPM_KEY_PARMS *parms)
{
  return FALSE;
}


/**********************************************************************
 * .._parms_is_good_es returns true is the ES is a strong value (can
 * deprecate use of weaker ESes by only recognizing stronger schemes
 * as "good"
 */
static BOOL null_parms_is_good_es(TPM_KEY_PARMS *parms)
{
  return FALSE;
}


/**********************************************************************
 * .._parms_is_any_es returns true if ES is set to any valid encryption
 * scheme for this keytype/algorithm
 */
static BOOL null_parms_is_any_es(TPM_KEY_PARMS *parms)
{
  return FALSE;
}


/**********************************************************************
 * .._parms_is_good_ss returns true is the SS is a strong value (can
 * deprecate use of weaker SSes by only recognizing stronger schemes
 * as "good"
 */
static BOOL null_parms_is_good_ss(TPM_KEY_PARMS *parms)
{
  return FALSE;
}


/**********************************************************************
 * .._parms_is_idkey_ss returns true iff this SS is a valid signature
 * scheme for an AIK to use
 */
static BOOL null_parms_is_idkey_ss(TPM_KEY_PARMS *parms)
{
  return FALSE;
}


/**********************************************************************
 * .._parms_is_any_ss returns true if SS is set to any valid signature
 * scheme for this keytype/algorithm
 */
static BOOL null_parms_is_any_ss(TPM_KEY_PARMS *parms)
{
  return FALSE;
}


/**********************************************************************
 * .._parms_requires_info returns true iff this signature scheme requires
 * the use of a TPM_SIGN_INFO structure
 */
static BOOL null_parms_requires_info(TPM_KEY_PARMS *parms)
{
  return FALSE;
}


/**********************************************************************
 * .._key_is_valid_keylengths checks public and private key lengths and
 * returns true if they are compatible/agree with each other
 */
static BOOL null_key_is_valid_keylengths(TPM_KEY *key, TPM_STORE_ASYMKEY *store)
{
  return FALSE;
}


/**********************************************************************
 * .._parms_get_ctextsize returns the size of the ciphertext (in bytes)
 * produced by encryption with a key described in parms
 */
static UINT32 null_parms_get_ctextsize(TPM_KEY_PARMS *parms)
{
  return 0;
}


/**********************************************************************
 * .._parms_get_sigsize returns the size of the signature (in bytes)
 * produced by signing with a key described in parms
 */
static UINT32 null_key_parms_get_sigsize(TPM_KEY_PARMS *parms)
{
  return 0;
}


/**********************************************************************
 * .._parms_get_ptextsizemax returns the maximum size of a plaintext
 * (in bytes) supported by an encryption algorithm/scheme described
 * in parms
 */
static UINT32 null_parms_get_ptextsizemax(TPM_KEY_PARMS *parms)
{
  return 0;
}


/**********************************************************************
 * .._parms_set_strong_keylen set the keylength in parms to be a size
 * that supports strong security (e.g., 2048 bits for an RSA key)
 */
static void null_parms_set_strong_keylen(TPM_KEY_PARMS *parms)
{
}


/**********************************************************************
 * .._parms_set_strong_es sets the parms encryption scheme to be a scheme
 * that provides strong security
 */
static void null_parms_set_strong_es(TPM_KEY_PARMS *parms)
{
}


/**********************************************************************
 * .._parms_set_default_algparms sets the parms structure to default
 * values.  In the base RSA scheme, this uses a minimal keysize, but there's
 * no reason it really has to be done that way.
 */
static int null_parms_set_default_algparms(TPM_KEY_PARMS *parms)
{
  return -1;
}


/**********************************************************************
 * .._parms_init_algparms initializes the binary blob /
 * algorithm-specific parameters by copying from the first param (src
 * parms) - must allocate any extra storage needed by algparms
 */
static int null_parms_init_algparms(TPM_KEY_PARMS *src, BYTE *destalgparms)
{
  return -1;
}


/**********************************************************************
 * .._parms_native_init_storepubkey initializes the TPM_STORE_PUBKEY
 * structure from the input parms and native-format keypair - must
 * allocate any storage needed by the STORE_PUBKEY
 */
static int null_parms_native_init_storepubkey(TPM_KEY_PARMS *parms, 
					 BYTE *native, TPM_STORE_PUBKEY *out)
{
  return -1;
}
    

/**********************************************************************
 * .._parms_nativepk_init_storepubkey initializes the TPM_STORE_PUBKEY
 * structure from the input parms and nativepk-format public key - must
 * allocate any storage needed by the STORE_PUBKEY
 */
static int null_parms_nativepk_init_storepubkey(TPM_KEY_PARMS *parms, 
					 BYTE *nativepk, TPM_STORE_PUBKEY *out)
{
  return -1;
}


/**********************************************************************
 * .._parms_native_init_storeprivkey initializes the TPM_STORE_PRIVKEY
 * structure from the input parms and native-format keypair - must allocate
 * any storage needed by STORE_PRIVKEY
 */
static int null_parms_native_init_storeprivkey(TPM_KEY_PARMS *parms,
				   BYTE *nativeKey, TPM_STORE_PRIVKEY *privKey)
{
  return -1;
}


/**********************************************************************
 * .._parms_storekeys_init_native initializes native-format keypair
 * from the public and private keys passed as generic TPM_STORE_... structs
 */
static int null_parms_storekeys_init_native(TPM_KEY_PARMS *parms, 
				       TPM_STORE_PUBKEY *spubkey,
				       TPM_STORE_PRIVKEY *sprivkey,
				       BYTE *native)
{
  return -1;
}


/**********************************************************************
 * .._parms_native_init_native copies one native-format keypair
 * to another 
 */
static int null_parms_native_init_native(TPM_KEY_PARMS *parms, 
					 BYTE *nativein,
					 BYTE *nativeout)
{
  return -1;
}


/**********************************************************************
 * .._pubkey_init_nativepk initializes a native-format public key from
 * a generic TPM_PUBKEY structure
 */
static int null_pubkey_init_nativepk(TPM_PUBKEY *src, BYTE *nativepk)
{
  return -1;
}


/**********************************************************************
 * .._keydata_init_nativepk initializes a native-format public key from
 * a generic TPM_KEY_DATA structure
 */
static int null_keydata_init_nativepk(TPM_KEY_DATA *key, BYTE *nativepk)
{
  return -1;
}


/**********************************************************************
 * .._parms_generate_native generates a new random native-format keypair
 * according to parameters passed in using the parms structure
 */
static int null_parms_generate_native(TPM_KEY_PARMS *parms, BYTE *native)
{
  return -1;
}


/**********************************************************************
 * .._pubkeydata_encrypt does an encryption in the obvious way (all buffers
 * must be pre-allocated)
 */
static int null_pubkeydata_encrypt(TPM_PUBKEY_DATA *key, BYTE *in,
				   UINT32 in_size, BYTE *enc, UINT32 *enc_size)
{
  return -1;
}


/**********************************************************************
 * .._keydata_decrypt does a decryption in the obvious way (all buffers
 * must be pre-allocated)
 */
static int null_keydata_decrypt(TPM_KEY_DATA *key, BYTE *enc, UINT32 enc_size,
                BYTE *out, UINT32 *out_size)
{
  return -1;
}


/**********************************************************************
 * .._parms_native_sign does a signature in the obvious way (all
 * buffers must be pre-allocated) - usually a hash is passed in, but
 * if instead raw data is passed in that should be hashed first you
 * can set the "hashFirst" flag
 */
static TPM_RESULT null_parms_native_sign(TPM_KEY_PARMS *parms, BYTE *nativeKey, 
	    const uint8_t *data, size_t data_len, BOOL hashFirst, uint8_t *sig)
{
  return TPM_FAIL;
}


/**********************************************************************
 * .._parms_nativepk_verify verifies a signature using the supplied public
 * key - usually a hash is passed in, but if instead raw data is passed in
 * that should be hashed first you can set the "hashFirst"  flag
 */
static TPM_RESULT null_parms_nativepk_verify(TPM_KEY_PARMS *parms, 
        BYTE *nativePK, BYTE *data, UINT32 data_len, BOOL hashFirst, BYTE *sig)
{
  return TPM_FAIL;
}


/**********************************************************************
 * .._parms_native_marshal_size returns the size (in bytes) that is required
 * when a native-format keypair is marshalled.
 */
static UINT32 null_parms_native_marshal_size(TPM_KEY_PARMS *parms, BYTE *native)
{
  return 0;
}


/**********************************************************************
 * .._parms_marshal_native marshals the provided native-format keypair into
 * a generic binary blob
 */
static int null_parms_marshal_native(TPM_KEY_PARMS *parms, BYTE **ptr,
			      UINT32 *length, BYTE *native)
{
  return -1;
}


/**********************************************************************
 * .._parms_unmarshal_native unmarshals the provided blob into
 * a native-format keypair
 */
static int null_parms_unmarshal_native(TPM_KEY_PARMS *parms, BYTE **ptr, 
				UINT32 *length, BYTE *native)
{
  return -1;
}


/**********************************************************************
 * .._parms_nativepk_marshal_size returns the size (in bytes) that is required
 * when a native-format public key is marshalled.
 */
static UINT32 null_parms_nativepk_marshal_size(TPM_KEY_PARMS *parms,
					       BYTE *nativepk)
{
  return 0;
}


/**********************************************************************
 * .._parms_marshal_nativepk marshals the provided native-format public key
 * into a generic binary blob
 */
static int null_parms_marshal_nativepk(TPM_KEY_PARMS *parms, BYTE **ptr,
			      UINT32 *length, BYTE *nativepk)
{
  return -1;
}


/**********************************************************************
 * .._parms_unmarshal_nativepk unmarshals the provided blob into
 * a native-format public key
 */
static int null_parms_unmarshal_nativepk(TPM_KEY_PARMS *parms, BYTE **ptr, 
				UINT32 *length, BYTE *nativepk)
{
  return -1;
}


/**********************************************************************
 * .._parms_marshal_algkeyparms marshals the algorithm-specific portion of
 * the keyparms
 */
static int null_parms_marshal_algkeyparms(TPM_KEY_PARMS *parms, BYTE **ptr,
				   UINT32 *length)
{
  return -1;
}


/**********************************************************************
 * .._parms_unmarshal_algkeyparms unmarshals into the algorithm-specific
 * portion of the keyparms
 */
static int null_parms_unmarshal_algkeyparms(TPM_KEY_PARMS *parms, BYTE **ptr,
				     UINT32 *length)
{
  return -1;
}


/**********************************************************************
 * .._parms_free frees up any dynamically-allocated space used in the
 * algorithm-specific keyparms
 */
static void null_parms_free(TPM_KEY_PARMS *parms)
{
}


/**********************************************************************
 * .._parms_native_free frees up any dynamically-allocated space used in the
 * native-format keypair
 */
static void null_parms_native_free(TPM_KEY_PARMS *parms, BYTE *native)
{
}


/**********************************************************************
 * .._parms_nativepk_free frees up any dynamically-allocated space used
 * in the native-format public key
 */
static void null_parms_nativepk_free(TPM_KEY_PARMS *parms, BYTE *pnativepk)
{
}


/**********************************************************************
 * .._selftest runs consistency checks on the routines for this keytype
 */
static int null_selftest(void)
{
  return 0;  // Success: null keys are always "functional"...  (?)
}


/**********************************************************************
 * .._init initializes any necessary data for this keytype, and makes sure
 * that keytype can be supported in the binary blocks that are allocated
 * for parms and native key representations
 */
static int null_init(void)
{
  return 0;
}


/**********************************************************************
 * This is the keytype structure that gives pointers to necessary routines
 * for this keytype
 */
keytype_info null_info = {
  0,   /* algorithmID */
  "NULL",
  null_init,

  null_is_genkey,
  null_is_transkey,
  null_es_is_validtrans,

  null_parms_is_supported,
  null_parms_is_minimal_size,
  null_parms_is_fips_size,
  null_parms_is_strong_size,
  null_parms_is_good_es,
  null_parms_is_any_es,
  null_parms_is_good_ss,
  null_parms_is_idkey_ss,
  null_parms_is_any_ss,
  null_parms_requires_info,
  null_key_is_valid_keylengths,

  null_parms_get_ctextsize,
  null_key_parms_get_sigsize,
  null_parms_get_ptextsizemax,

  null_parms_set_strong_keylen,
  null_parms_set_strong_es,

  null_parms_set_default_algparms,
  null_parms_init_algparms,

  null_parms_native_init_storepubkey,
  null_parms_nativepk_init_storepubkey,

  null_parms_native_init_storeprivkey,

  null_parms_storekeys_init_native,
  null_parms_native_init_native,

  null_pubkey_init_nativepk,
  null_keydata_init_nativepk,

  null_parms_generate_native,
  null_pubkeydata_encrypt,
  null_keydata_decrypt,
  null_parms_native_sign,
  null_parms_nativepk_verify,

  null_parms_native_marshal_size,
  null_parms_marshal_native,
  null_parms_unmarshal_native,

  null_parms_nativepk_marshal_size,
  null_parms_marshal_nativepk,
  null_parms_unmarshal_nativepk,

  null_parms_marshal_algkeyparms,
  null_parms_unmarshal_algkeyparms,

  null_parms_free,
  null_parms_native_free,
  null_parms_nativepk_free,

  null_selftest
};
