/* Software-based Trusted Platform Module (TPM) Emulator
 * Emulator Copyright (C) 2004-2010 Mario Strasser <mast@gmx.net>
 * Timing/plugin code Copyright (C) 2010-2012 Stephen R. Tate <srtate@uncg.edu>
 *
 * This module is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 *
 * This module is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 */

#include "crypto/sha1.h"
#include "crypto/rsa.h"
#include "tpm/tpm_marshalling.h"
#include "tpm/plugins.h"

/*                      */
/*    RSA functions     */
/*                      */


/**********************************************************************
 * .._is_genkey() means this keytype/algorithm can be used for Sign, Seal, 
 * Unbind, etc. operations
 */
static BOOL rsa_is_genkey(void)
{
  return TRUE;
}


/**********************************************************************
 * .._is_transkey() means this keytype/algorithm can be used for
 * EstablishTransport, ExecuteTransport, and ReleaseTransportSigned
 */
static BOOL rsa_is_transkey(void)
{
  return TRUE;
}


/**********************************************************************
 * .._es_is_validtrans returns true is this is a valid encryption scheme (ES)
 * for a transport session.  Actually only used when getting capability
 * TPM_CAP_TRANS_ES
 */
static BOOL rsa_es_is_validtrans(TPM_ENC_SCHEME es)
{
  switch (es) {
    case TPM_ES_RSAESOAEP_SHA1_MGF1:
    case TPM_ES_RSAESPKCSv15:
      return TRUE;
    default:
      return FALSE;
  }
}


/**********************************************************************
 * .._parms_is_supported returns true if the parameters in the parms
 * structure supported?  (key size, etc.)
 */
static BOOL rsa_parms_is_supported(TPM_KEY_PARMS *parms)
{
  TPM_RSA_KEY_PARMS *rsa = (TPM_RSA_KEY_PARMS *)parms->bparms;
  return (rsa->keyLength <= 2048)
    && (rsa->numPrimes == 2)
    && (rsa->exponentSize == 0);
}


/**********************************************************************
 * .._parms_is_minimal_size returns true if the keysize is at least some
 * minimal reasonable value
 */
static BOOL rsa_parms_is_minimal_size(TPM_KEY_PARMS *parms)
{
  TPM_RSA_KEY_PARMS *rsa = (TPM_RSA_KEY_PARMS *)parms->bparms;
  return rsa->keyLength >= 512;
}


/**********************************************************************
 * .._parms_is_fips_size returns true if the keysize is at least some
 * the size required by FIPS requirements (maybe be more than the minimal
 * size otherwise supported).
 */
static BOOL rsa_parms_is_fips_size(TPM_KEY_PARMS *parms)
{
  TPM_RSA_KEY_PARMS *rsa = (TPM_RSA_KEY_PARMS *)parms->bparms;
  return rsa->keyLength >= 1024;
}


/**********************************************************************
 * .._parms_is_strong_size returns true if the keysize is sufficient to
 * support strong security (e.g., 2048 bits for an RSA key)
 */
static BOOL rsa_parms_is_strong_size(TPM_KEY_PARMS *parms)
{
  TPM_RSA_KEY_PARMS *rsa = (TPM_RSA_KEY_PARMS *)parms->bparms;
  return rsa->keyLength == 2048;
}


/**********************************************************************
 * .._parms_is_good_es returns true is the ES is a strong value (can
 * deprecate use of weaker ESes by only recognizing stronger schemes
 * as "good"
 */
static BOOL rsa_parms_is_good_es(TPM_KEY_PARMS *parms)
{
  return (parms->encScheme == TPM_ES_RSAESOAEP_SHA1_MGF1) 
    && (parms->sigScheme == TPM_SS_NONE);
}


/**********************************************************************
 * .._parms_is_any_es returns true if ES is set to any valid encryption
 * scheme for this keytype/algorithm
 */
static BOOL rsa_parms_is_any_es(TPM_KEY_PARMS *parms)
{
  return (parms->encScheme != TPM_ES_NONE) 
    && (parms->sigScheme == TPM_SS_NONE);
}


/**********************************************************************
 * .._parms_is_good_ss returns true is the SS is a strong value (can
 * deprecate use of weaker SSes by only recognizing stronger schemes
 * as "good"
 */
static BOOL rsa_parms_is_good_ss(TPM_KEY_PARMS *parms)
{
  return (parms->encScheme == TPM_ES_NONE) 
    && ((parms->sigScheme == TPM_SS_RSASSAPKCS1v15_SHA1)
	|| (parms->sigScheme == TPM_SS_RSASSAPKCS1v15_INFO));
}


/**********************************************************************
 * .._parms_is_idkey_ss returns true iff this SS is a valid signature
 * scheme for an AIK to use
 */
static BOOL rsa_parms_is_idkey_ss(TPM_KEY_PARMS *parms)
{
  return (parms->encScheme == TPM_ES_NONE) 
    && (parms->sigScheme == TPM_SS_RSASSAPKCS1v15_SHA1);
}


/**********************************************************************
 * .._parms_is_any_ss returns true if SS is set to any valid signature
 * scheme for this keytype/algorithm
 */
static BOOL rsa_parms_is_any_ss(TPM_KEY_PARMS *parms)
{
  return (parms->encScheme == TPM_ES_NONE) 
    && (parms->sigScheme != TPM_SS_NONE);
}


/**********************************************************************
 * .._parms_requires_info returns true iff this signature scheme requires
 * the use of a TPM_SIGN_INFO structure
 */
static BOOL rsa_parms_requires_info(TPM_KEY_PARMS *parms)
{
  return parms->sigScheme == TPM_SS_RSASSAPKCS1v15_INFO;
}


/**********************************************************************
 * .._key_is_valid_keylengths checks public and private key lengths and
 * returns true if they are compatible/agree with each other
 */
static BOOL rsa_key_is_valid_keylengths(TPM_KEY *key, TPM_STORE_ASYMKEY *store)
{
  return key->pubKey.keyLength == 2*store->privKey.keyLength;
}


/**********************************************************************
 * .._parms_get_ctextsize returns the size of the ciphertext (in bytes)
 * produced by encryption with a key described in parms
 */
static UINT32 rsa_parms_get_ctextsize(TPM_KEY_PARMS *parms)
{
  TPM_RSA_KEY_PARMS *rsa = (TPM_RSA_KEY_PARMS *)parms->bparms;
  return (rsa->keyLength+7)/8;
}


/**********************************************************************
 * .._parms_get_sigsize returns the size of the signature (in bytes)
 * produced by signing with a key described in parms
 */
static UINT32 rsa_key_parms_get_sigsize(TPM_KEY_PARMS *parms)
{
  TPM_RSA_KEY_PARMS *rsa = (TPM_RSA_KEY_PARMS *)parms->bparms;
  return (rsa->keyLength+7)/8;
}


/**********************************************************************
 * .._parms_get_ptextsizemax returns the maximum size of a plaintext
 * (in bytes) supported by an encryption algorithm/scheme described
 * in parms
 */
static UINT32 rsa_parms_get_ptextsizemax(TPM_KEY_PARMS *parms)
{
  TPM_RSA_KEY_PARMS *rsa = (TPM_RSA_KEY_PARMS *)parms->bparms;
  switch (parms->encScheme) {
  case TPM_ES_RSAESPKCSv15:
    return ((rsa->keyLength+7)>>3)-11;
  case TPM_ES_RSAESOAEP_SHA1_MGF1:
    return ((rsa->keyLength+7)>>3) - 2*SHA1_DIGEST_LENGTH - 2;
  }
  return 0;
}


/**********************************************************************
 * .._parms_set_strong_keylen set the keylength in parms to be a size
 * that supports strong security (e.g., 2048 bits for an RSA key)
 */
static void rsa_parms_set_strong_keylen(TPM_KEY_PARMS *parms)
{
  TPM_RSA_KEY_PARMS *rsa = (TPM_RSA_KEY_PARMS *)parms->bparms;
  rsa->keyLength = 2048;
}


/**********************************************************************
 * .._parms_set_strong_es sets the parms encryption scheme to be a scheme
 * that provides strong security
 */
static void rsa_parms_set_strong_es(TPM_KEY_PARMS *parms)
{
  parms->encScheme = TPM_ES_RSAESOAEP_SHA1_MGF1;
  parms->sigScheme = TPM_SS_NONE;
}


/**********************************************************************
 * .._parms_set_default_algparms sets the parms structure to default
 * values.  In the base RSA scheme, this uses a minimal keysize, but there's
 * no reason it really has to be done that way.
 */
static int rsa_parms_set_default_algparms(TPM_KEY_PARMS *parms)
{
  TPM_RSA_KEY_PARMS *rsa = (TPM_RSA_KEY_PARMS *)parms->bparms;
  parms->parmSize = 12;
  rsa->keyLength = 512;
  rsa->numPrimes = 2;
  rsa->exponentSize = 0;
  rsa->exponent = NULL;
  return 0;
}


/**********************************************************************
 * .._parms_init_algparms initializes the binary blob /
 * algorithm-specific parameters by copying from the first param (src
 * parms) - must allocate any extra storage needed by algparms
 */
static int rsa_parms_init_algparms(TPM_KEY_PARMS *src, BYTE *destalgparms)
{
  TPM_RSA_KEY_PARMS *srcrsa = (TPM_RSA_KEY_PARMS *)src->bparms;
  TPM_RSA_KEY_PARMS *destrsa = (TPM_RSA_KEY_PARMS *)destalgparms;

  destrsa->keyLength = srcrsa->keyLength;
  destrsa->numPrimes = srcrsa->numPrimes;
  destrsa->exponentSize = srcrsa->exponentSize;
  if (srcrsa->exponentSize == 0) {
    destrsa->exponent = NULL;
  } else {
    destrsa->exponent = tpm_malloc(srcrsa->exponentSize);
    if (destrsa->exponent == NULL) {
      destrsa->exponentSize = 0;
      return -1;
    }
    memcpy(destrsa->exponent, srcrsa->exponent, srcrsa->exponentSize);
  }
  return 0;
}


/**********************************************************************
 * .._parms_native_init_storepubkey initializes the TPM_STORE_PUBKEY
 * structure from the input parms and native-format keypair - must
 * allocate any storage needed by the STORE_PUBKEY
 */
static int rsa_parms_native_init_storepubkey(TPM_KEY_PARMS *src, BYTE *native,
				      TPM_STORE_PUBKEY *out)
{
    tpm_rsa_private_key_t *rsakey = (tpm_rsa_private_key_t *)native;
    UINT32 key_length = (rsakey->size+7)>>3;
    out->key = tpm_malloc(key_length);
    if (out->key == NULL) {
	out->keyLength = 0;
	return -1;
    }
    tpm_rsa_export_modulus(rsakey, out->key, &key_length);
    out->keyLength = key_length;
    return 0;
}


/**********************************************************************
 * .._parms_nativepk_init_storepubkey initializes the TPM_STORE_PUBKEY
 * structure from the input parms and nativepk-format public key - must
 * allocate any storage needed by the STORE_PUBKEY
 */
static int rsa_parms_nativepk_init_storepubkey(TPM_KEY_PARMS *parms,
					       BYTE *nativepk,
					       TPM_STORE_PUBKEY *out)
{
    tpm_rsa_public_key_t *rsapub = (tpm_rsa_public_key_t *)nativepk;
    size_t key_length = (rsapub->size+7) >> 3;
    out->key = tpm_malloc(key_length);
    if (out->key == NULL) {
	out->keyLength = 0;
	return -1;
    }
    tpm_rsa_export_public_modulus(rsapub, out->key, &key_length);
    out->keyLength = key_length;
    return 0;
}


/**********************************************************************
 * .._parms_native_init_storeprivkey initializes the TPM_STORE_PRIVKEY
 * structure from the input parms and native-format keypair - must allocate
 * any storage needed by STORE_PRIVKEY
 */
static int rsa_parms_native_init_storeprivkey(TPM_KEY_PARMS *parms,
					      BYTE *nativeKey,
					      TPM_STORE_PRIVKEY *privKey)
{
    tpm_rsa_private_key_t *rsakey = (tpm_rsa_private_key_t *)nativeKey;
    privKey->keyLength = (rsakey->size+7)/16;
    privKey->key = tpm_malloc(privKey->keyLength);
    if (privKey->key == NULL) {
	privKey->keyLength = 0;
	return -1;
    }
    tpm_rsa_export_prime1(rsakey, privKey->key, NULL);
    return 0;
}


/**********************************************************************
 * .._parms_storekeys_init_native initializes native-format keypair
 * from the public and private keys passed as generic TPM_STORE_... structs
 */
static int rsa_parms_storekeys_init_native(TPM_KEY_PARMS *parms, 
				      TPM_STORE_PUBKEY *spubkey,
				      TPM_STORE_PRIVKEY *sprivkey,
				      BYTE *native)
{
  TPM_RSA_KEY_PARMS *rsaparms = (TPM_RSA_KEY_PARMS *)parms->bparms;
  tpm_rsa_private_key_t *rsapriv = (tpm_rsa_private_key_t *)native;
  return tpm_rsa_import_key(rsapriv, RSA_MSB_FIRST, spubkey->key, 
			    spubkey->keyLength, rsaparms->exponent,
			    rsaparms->exponentSize, sprivkey->key, NULL);
}


/**********************************************************************
 * .._parms_native_init_native copies one native-format keypair
 * to another 
 */
static int rsa_parms_native_init_native(TPM_KEY_PARMS *parms, 
					BYTE *nativein,
					BYTE *nativeout)
{
  tpm_rsa_private_key_t *rsain = (tpm_rsa_private_key_t *)nativein;
  tpm_rsa_private_key_t *rsaout = (tpm_rsa_private_key_t *)nativein;
  tpm_rsa_copy_key(rsaout, rsain);
  return 0;
}


/**********************************************************************
 * .._pubkey_init_nativepk initializes a native-format public key from
 * a generic TPM_PUBKEY structure
 */
static int rsa_pubkey_init_nativepk(TPM_PUBKEY *src, BYTE *nativepk)
{
  TPM_RSA_KEY_PARMS *rsa = (TPM_RSA_KEY_PARMS *)src->algorithmParms.bparms;
  tpm_rsa_public_key_t *rsapubkey = (tpm_rsa_public_key_t *)nativepk;
  return tpm_rsa_import_public_key(rsapubkey, RSA_MSB_FIRST, src->pubKey.key,
				   src->pubKey.keyLength, rsa->exponent,
				   rsa->exponentSize);
}


/**********************************************************************
 * .._keydata_init_nativepk initializes a native-format public key from
 * a generic TPM_KEY_DATA structure
 */
static int rsa_keydata_init_nativepk(TPM_KEY_DATA *key, BYTE *nativepk)
{
  tpm_rsa_private_key_t *rsakey = (tpm_rsa_private_key_t *)key->bnativekey;
  tpm_rsa_public_key_t *rsapub = (tpm_rsa_public_key_t *)nativepk;
  TPM_RSA_EXTRACT_PUBLIC_KEY((*rsakey), (*rsapub));
  return 0;
}


/**********************************************************************
 * .._parms_generate_native generates a new random native-format keypair
 * according to parameters passed in using the parms structure
 */
static int rsa_parms_generate_native(TPM_KEY_PARMS *parms, BYTE *nativeout)
{
  TPM_RSA_KEY_PARMS *rsa = (TPM_RSA_KEY_PARMS *)parms->bparms;
  tpm_rsa_private_key_t *native = (tpm_rsa_private_key_t *)nativeout;
  return tpm_rsa_generate_key(native, rsa->keyLength);
}


/**********************************************************************
 * .._pubkeydata_encrypt does an encryption in the obvious way (all buffers
 * must be pre-allocated)
 */
static int rsa_pubkeydata_encrypt(TPM_PUBKEY_DATA *key, BYTE *in,
				  UINT32 in_size, BYTE *enc, UINT32 *enc_size)
{
  tpm_rsa_public_key_t *rsapubkey = (tpm_rsa_public_key_t *)key->bnativepubkey;
  size_t size = *enc_size;
  int scheme;
  switch (key->algorithmParms.encScheme) {
  case TPM_ES_RSAESOAEP_SHA1_MGF1: scheme = RSA_ES_OAEP_SHA1; break;
  case TPM_ES_RSAESPKCSv15: scheme = RSA_ES_PKCSV15; break;
  default:
    debug("unsupported encryption scheme: %d", key->algorithmParms.encScheme);
    return -1;
  }
  if (tpm_rsa_encrypt(rsapubkey, scheme, in, in_size, enc, &size) != 0) {
    debug("tpm_rsa_encrypt() failed");
    return -1;
  }
  *enc_size = size;
  return 0;
}


/**********************************************************************
 * .._keydata_decrypt does a decryption in the obvious way (all buffers
 * must be pre-allocated)
 */
static int rsa_keydata_decrypt(TPM_KEY_DATA *key, BYTE *enc, UINT32 enc_size,
			BYTE *out, UINT32 *out_size)
{
  tpm_rsa_private_key_t *rsakey = (tpm_rsa_private_key_t *)key->bnativekey;
  size_t size = *out_size;
  int scheme;
  switch (key->algorithmParms.encScheme) {
    case TPM_ES_RSAESOAEP_SHA1_MGF1: scheme = RSA_ES_OAEP_SHA1; break;
    case TPM_ES_RSAESPKCSv15: scheme = RSA_ES_PKCSV15; break;
    default:
      debug("unsupported encryption scheme: %d", key->algorithmParms.encScheme);
      return -1;
  }
  if (tpm_rsa_decrypt(rsakey, scheme, enc, enc_size, out, &size) != 0) {
    debug("tpm_rsa_decrypt() failed");
    return -1;
  }
  *out_size = size;
  return 0;
}


/**********************************************************************
 * .._parms_native_sign does a signature in the obvious way (all
 * buffers must be pre-allocated) - usually a hash is passed in, but
 * if instead raw data is passed in that should be hashed first you
 * can set the "hashFirst" flag
 */
static TPM_RESULT rsa_parms_native_sign(TPM_KEY_PARMS *parms, BYTE *nativeKey, 
           const uint8_t *data, size_t data_len, BOOL hashFirst, uint8_t *sig)
{
  TPM_RSA_KEY_PARMS *rsa = (TPM_RSA_KEY_PARMS *)parms->bparms;
  tpm_rsa_private_key_t *rsakey = (tpm_rsa_private_key_t *)nativeKey;
  int scheme = 0;
  switch (parms->sigScheme) {
  case TPM_SS_RSASSAPKCS1v15_SHA1:
  case TPM_SS_RSASSAPKCS1v15_INFO:
    scheme = RSA_SSA_PKCS1_SHA1;
    break;
  case TPM_SS_RSASSAPKCS1v15_DER:
    if (data_len == 0 || data_len+11 > rsa->keyLength)
      return TPM_BAD_PARAMETER;
    scheme = RSA_SSA_PKCS1_DER;
    break;
  default:
    debug("rsa_parms_native_sign(): Invalid sigScheme: %d\n", parms->sigScheme);
    return TPM_INVALID_KEYUSAGE;
  }

  if (scheme == RSA_SSA_PKCS1_SHA1 && !hashFirst) {
    if (data_len != 20) return TPM_BAD_PARAMETER;
    scheme = RSA_SSA_PKCS1_SHA1_RAW;
  }

  if (tpm_rsa_sign(rsakey, scheme, data, data_len, sig))
    return TPM_FAIL;
  return TPM_SUCCESS;
}


/**********************************************************************
 * .._parms_nativepk_verify verifies a signature using the supplied public
 * key - usually a hash is passed in, but if instead raw data is passed in
 * that should be hashed first you can set the "hashFirst"  flag
 */
static TPM_RESULT rsa_parms_nativepk_verify(TPM_KEY_PARMS *parms,
	BYTE *nativePK, BYTE *data, UINT32 data_len, BOOL hashFirst, BYTE *sig)
{
  TPM_RSA_KEY_PARMS *rsa = (TPM_RSA_KEY_PARMS *)parms->bparms;
  tpm_rsa_public_key_t *rsakey = (tpm_rsa_public_key_t *)nativePK;
  int scheme = 0;
  switch (parms->sigScheme) {
  case TPM_SS_RSASSAPKCS1v15_SHA1:
  case TPM_SS_RSASSAPKCS1v15_INFO:
    scheme = RSA_SSA_PKCS1_SHA1;
    break;
  case TPM_SS_RSASSAPKCS1v15_DER:
    if (data_len == 0 || data_len+11 > rsa->keyLength)
      return TPM_BAD_PARAMETER;
    scheme = RSA_SSA_PKCS1_DER;
    break;
  default:
    debug("rsa_parms_native_sign(): Invalid sigScheme: %d\n", parms->sigScheme);
    return TPM_INVALID_KEYUSAGE;
  }

  if (scheme == RSA_SSA_PKCS1_SHA1 && !hashFirst)
    scheme = RSA_SSA_PKCS1_SHA1_RAW;

  if (tpm_rsa_verify(rsakey, scheme, data, data_len, sig))
    return TPM_BAD_SIGNATURE;
  return TPM_SUCCESS;
}


/**********************************************************************
 * .._parms_native_marshal_size returns the size (in bytes) that is required
 * when a native-format keypair is marshalled.
 */
static UINT32 rsa_parms_native_marshal_size(TPM_KEY_PARMS *parms, BYTE *native)
{
  tpm_rsa_private_key_t *rsakey = (tpm_rsa_private_key_t *)native;
  return 6 + tpm_rsa_modulus_length(rsakey) 
    + tpm_rsa_exponent_length(rsakey) + tpm_rsa_prime1_length(rsakey);
}


/**********************************************************************
 * .._parms_marshal_native marshals the provided native-format keypair into
 * a generic binary blob
 */
static int rsa_parms_marshal_native(TPM_KEY_PARMS *parms, BYTE **ptr, 
			     UINT32 *length, BYTE *native)
{
  tpm_rsa_private_key_t *v = (tpm_rsa_private_key_t *)native;
  size_t m_len, e_len, q_len;
  if (*length < rsa_parms_native_marshal_size(parms, native)) return -1;
  if (v->size > 0) {
    tpm_rsa_export_modulus(v, &(*ptr)[6], &m_len);
    tpm_rsa_export_exponent(v, &(*ptr)[6+m_len], &e_len);
    tpm_rsa_export_prime1(v, &(*ptr)[6+m_len+e_len], &q_len);
    tpm_marshal_UINT16(ptr, length, m_len);
    tpm_marshal_UINT16(ptr, length, e_len);
    tpm_marshal_UINT16(ptr, length, q_len);
    *ptr += m_len + e_len + q_len;
    *length -= m_len + e_len + q_len;
  } else {
    tpm_marshal_UINT16(ptr, length, 0);
    tpm_marshal_UINT16(ptr, length, 0);
    tpm_marshal_UINT16(ptr, length, 0);
  }
  return 0;
}


/**********************************************************************
 * .._parms_unmarshal_native unmarshals the provided blob into
 * a native-format keypair
 */
static int rsa_parms_unmarshal_native(TPM_KEY_PARMS *parms, BYTE **ptr,
			       UINT32 *length, BYTE *native)
{
  tpm_rsa_private_key_t *rsakey = (tpm_rsa_private_key_t *)native;

  UINT16 m_len, e_len, q_len;
  if (tpm_unmarshal_UINT16(ptr, length, &m_len)
      || tpm_unmarshal_UINT16(ptr, length, &e_len)
      || tpm_unmarshal_UINT16(ptr, length, &q_len)) return -1;
  if (m_len == 0) {
    rsakey->size = 0;
    return 0;
  }
  if (*length < (UINT32)m_len + (UINT32)e_len + (UINT32)q_len
      || q_len != m_len/2
      || tpm_rsa_import_key(rsakey, RSA_MSB_FIRST,
                        &(*ptr)[0], m_len,
                        &(*ptr)[m_len], e_len,
                        &(*ptr)[m_len+e_len], NULL)) return -1;
  *ptr += m_len + e_len + q_len;
  *length -= m_len + e_len + q_len;
  return 0;
}


/**********************************************************************
 * .._parms_nativepk_marshal_size returns the size (in bytes) that is required
 * when a native-format public key is marshalled.
 */
static UINT32 rsa_parms_nativepk_marshal_size(TPM_KEY_PARMS *parms,
					      BYTE *nativepk)
{
  tpm_rsa_public_key_t *rsakey = (tpm_rsa_public_key_t *)nativepk;
  return 4 + tpm_rsa_public_modulus_length(rsakey) 
    + tpm_rsa_public_exponent_length(rsakey);
}


/**********************************************************************
 * .._parms_marshal_nativepk marshals the provided native-format public key
 * into a generic binary blob
 */
static int rsa_parms_marshal_nativepk(TPM_KEY_PARMS *parms, BYTE **ptr, 
			     UINT32 *length, BYTE *nativepk)
{
  tpm_rsa_public_key_t *v = (tpm_rsa_public_key_t *)nativepk;
  size_t m_len, e_len;
  if (*length < rsa_parms_nativepk_marshal_size(parms, nativepk)) return -1;
  if (v->size > 0) {
    tpm_rsa_export_public_modulus(v, &(*ptr)[4], &m_len);
    tpm_rsa_export_public_exponent(v, &(*ptr)[4+m_len], &e_len);
    tpm_marshal_UINT16(ptr, length, m_len);
    tpm_marshal_UINT16(ptr, length, e_len);
    *ptr += m_len + e_len;
    *length -= m_len + e_len;
  } else {
    tpm_marshal_UINT16(ptr, length, 0);
    tpm_marshal_UINT16(ptr, length, 0);
  }
  return 0;
}


/**********************************************************************
 * .._parms_unmarshal_nativepk unmarshals the provided blob into
 * a native-format public key
 */
static int rsa_parms_unmarshal_nativepk(TPM_KEY_PARMS *parms, BYTE **ptr,
			       UINT32 *length, BYTE *nativepk)
{
  tpm_rsa_public_key_t *rsakey = (tpm_rsa_public_key_t *)nativepk;

  UINT16 m_len, e_len;
  if (tpm_unmarshal_UINT16(ptr, length, &m_len)
      || tpm_unmarshal_UINT16(ptr, length, &e_len)) return -1;
  if (m_len == 0) {
    rsakey->size = 0;
    return 0;
  }
  if (*length < (UINT32)m_len + (UINT32)e_len
      || tpm_rsa_import_public_key(rsakey, RSA_MSB_FIRST, &(*ptr)[0], m_len, 
                                   &(*ptr)[m_len], e_len)) return -1;
  *ptr += m_len + e_len;
  *length -= m_len + e_len;
  return 0;
}


/**********************************************************************
 * .._parms_marshal_algkeyparms marshals the algorithm-specific portion of
 * the keyparms
 */
static int rsa_parms_marshal_algkeyparms(TPM_KEY_PARMS *parms, BYTE **ptr,
				  UINT32 *length)
{
  TPM_RSA_KEY_PARMS *rsa = (TPM_RSA_KEY_PARMS *)parms->bparms;
  return tpm_marshal_TPM_RSA_KEY_PARMS(ptr, length, rsa);
}


/**********************************************************************
 * .._parms_unmarshal_algkeyparms unmarshals into the algorithm-specific
 * portion of the keyparms
 */
static int rsa_parms_unmarshal_algkeyparms(TPM_KEY_PARMS *parms, BYTE **ptr,
				    UINT32 *length)
{
  TPM_RSA_KEY_PARMS *rsa = (TPM_RSA_KEY_PARMS *)parms->bparms;
  return tpm_unmarshal_TPM_RSA_KEY_PARMS(ptr, length, rsa);
}


/**********************************************************************
 * .._parms_free frees up any dynamically-allocated space used in the
 * algorithm-specific keyparms
 */
static void rsa_parms_free(TPM_KEY_PARMS *parms)
{
  TPM_RSA_KEY_PARMS *rsa = (TPM_RSA_KEY_PARMS *)parms->bparms;
  if (rsa != NULL) {
    if (rsa->exponentSize > 0) {
      tpm_free(rsa->exponent);
      rsa->exponent = NULL;
      rsa->exponentSize = 0;
    }
  }
}


/**********************************************************************
 * .._parms_native_free frees up any dynamically-allocated space used in the
 * native-format keypair
 */
static void rsa_parms_native_free(TPM_KEY_PARMS *parms, BYTE *native)
{
  tpm_rsa_private_key_t *rsakey = (tpm_rsa_private_key_t *)native;
  tpm_rsa_release_private_key(rsakey);
}


/**********************************************************************
 * .._parms_nativepk_free frees up any dynamically-allocated space used
 * in the native-format public key
 */
static void rsa_parms_nativepk_free(TPM_KEY_PARMS *parms, BYTE *nativepk)
{
  tpm_rsa_public_key_t *rsapubkey = (tpm_rsa_public_key_t *)nativepk;
  tpm_rsa_release_public_key(rsapubkey);
}


/**********************************************************************
 * .._selftest runs consistency checks on the routines for this keytype
 */
static int rsa_selftest(void)
{
  int res = 0;
  uint8_t *data = (uint8_t*)"RSA PKCS #1 v1.5 Test-String";
  uint8_t buf[256];
  size_t buf_len, data_len = strlen((char*)data);
  tpm_rsa_private_key_t priv_key;
  tpm_rsa_public_key_t pub_key;

  debug("rsa_selftest()");
  /* generate and test key-pair */
  debug("tpm_rsa_generate_key()");
  res = tpm_rsa_generate_key(&priv_key, 2048);
  if (res) return res;
  /* test endorsement key */
  do {
    if (!priv_key.size) return 0;
    TPM_RSA_EXTRACT_PUBLIC_KEY(priv_key, pub_key);
    /* test sign and verify functions */
    debug("tpm_rsa_sign(RSA_SSA_PKCS1_SHA1)");
    res = tpm_rsa_sign(&priv_key, RSA_SSA_PKCS1_SHA1, data, data_len, buf);
    if (res) break;
    debug("tpm_rsa_verify(RSA_SSA_PKCS1_SHA1)");
    res = tpm_rsa_verify(&pub_key, RSA_SSA_PKCS1_SHA1, data, data_len, buf);
    if (res) break;
    debug("tpm_rsa_sign(RSA_SSA_PKCS1_DER)");
    res = tpm_rsa_sign(&priv_key, RSA_SSA_PKCS1_DER, data, data_len, buf);
    if (res) break;
    debug("tpm_rsa_verify(RSA_SSA_PKCS1_DER)");
    res = tpm_rsa_verify(&pub_key, RSA_SSA_PKCS1_DER, data, data_len, buf);
    if (res) break;
    /* test encryption and decryption */
    debug("tpm_rsa_encrypt(RSA_ES_PKCSV15)");
    res = tpm_rsa_encrypt(&pub_key, RSA_ES_PKCSV15,
      data, data_len, buf, &buf_len);
    if (res) break;
    debug("tpm_rsa_decrypt(RSA_ES_PKCSV15)");
    res = tpm_rsa_decrypt(&priv_key, RSA_ES_PKCSV15,
      buf, buf_len, buf, &buf_len);
    if (res) break;
    debug("verify plain text");
    res = !((buf_len == data_len) && !memcmp(buf, data, buf_len));
    if (res) break;
    debug("tpm_rsa_encrypt(RSA_ES_OAEP_SHA1)");
    res = tpm_rsa_encrypt(&pub_key, RSA_ES_OAEP_SHA1,
      data, data_len/2, buf, &buf_len);
    if (res) break;
    debug("tpm_rsa_decrypt(RSA_ES_OAEP_SHA1)");
    res = tpm_rsa_decrypt(&priv_key, RSA_ES_OAEP_SHA1,
      buf, buf_len, buf, &buf_len);
    if (res) break;
    debug("verify plain text");
    res = !(buf_len == data_len/2 && !memcmp(buf, data, buf_len));
  } while (0);
  /* release public key and exit */
  tpm_rsa_release_public_key(&pub_key);
  tpm_rsa_release_private_key(&priv_key);

  return res;
}


/**********************************************************************
 * .._init initializes any necessary data for this keytype, and makes sure
 * that keytype can be supported in the binary blocks that are allocated
 * for parms and native key representations
 */
int rsa_init(void)
{
  int ret = 0;
  if (sizeof(TPM_RSA_KEY_PARMS) > TPM_ALG_PARMS_MAXSIZE) {
    info("Can't load RSA keytype: need %d bytes for KEY_PARMS and only have %d",
	 sizeof(TPM_RSA_KEY_PARMS), TPM_ALG_PARMS_MAXSIZE);
    ret = -1;
  }
  if (sizeof(tpm_rsa_private_key_t) > TPM_NATIVE_KEY_MAXSIZE) {
    info("Can't load RSA keytype: need %d bytes for native and only have %d",
	 sizeof(tpm_rsa_private_key_t), TPM_NATIVE_KEY_MAXSIZE);
    ret = -1;
  }
  if (sizeof(tpm_rsa_public_key_t) > TPM_NATIVE_KEY_MAXSIZE) {
    info("Can't load RSA keytype: need %d bytes for nativepub and only have %d",
	 sizeof(tpm_rsa_public_key_t), TPM_NATIVE_KEY_MAXSIZE);
    ret = -1;
  }
  if (ret) {
    info("To make this work, you must recompile the TPM emulator with a bigger buffer");
  }
  return ret;
}


/**********************************************************************
 * This is the keytype structure that gives pointers to necessary routines
 * for this keytype
 */
keytype_info rsa_info = {
  TPM_ALG_RSA,
  "RSA",
  rsa_init,

  rsa_is_genkey,
  rsa_is_transkey,
  rsa_es_is_validtrans,

  rsa_parms_is_supported,
  rsa_parms_is_minimal_size,
  rsa_parms_is_fips_size,
  rsa_parms_is_strong_size,
  rsa_parms_is_good_es,
  rsa_parms_is_any_es,
  rsa_parms_is_good_ss,
  rsa_parms_is_idkey_ss,
  rsa_parms_is_any_ss,
  rsa_parms_requires_info,
  rsa_key_is_valid_keylengths,

  rsa_parms_get_ctextsize,
  rsa_key_parms_get_sigsize,
  rsa_parms_get_ptextsizemax,

  rsa_parms_set_strong_keylen,
  rsa_parms_set_strong_es,

  rsa_parms_set_default_algparms,
  rsa_parms_init_algparms,

  rsa_parms_native_init_storepubkey,
  rsa_parms_nativepk_init_storepubkey,

  rsa_parms_native_init_storeprivkey,

  rsa_parms_storekeys_init_native,
  rsa_parms_native_init_native,

  rsa_pubkey_init_nativepk,
  rsa_keydata_init_nativepk,

  rsa_parms_generate_native,
  rsa_pubkeydata_encrypt,
  rsa_keydata_decrypt,
  rsa_parms_native_sign,
  rsa_parms_nativepk_verify,

  rsa_parms_native_marshal_size,
  rsa_parms_marshal_native,
  rsa_parms_unmarshal_native,

  rsa_parms_nativepk_marshal_size,
  rsa_parms_marshal_nativepk,
  rsa_parms_unmarshal_nativepk,

  rsa_parms_marshal_algkeyparms,
  rsa_parms_unmarshal_algkeyparms,

  rsa_parms_free,
  rsa_parms_native_free,
  rsa_parms_nativepk_free,

  rsa_selftest
};
