/* Software-based Trusted Platform Module (TPM) Emulator
 * Emulator Copyright (C) 2004-2010 Mario Strasser <mast@gmx.net>
 * Timing/plugin code Copyright (C) 2010-2012 Stephen R. Tate <srtate@uncg.edu>
 *
 * This module is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 *
 * This module is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 */

/*
 * Original plugin/extension code by Alex Kilgore and Stephen Tate.
 */

#ifndef _KEY_FUNCTIONS_H
#define _KEY_FUNCTIONS_H

/*
 * Information about and definitions for TPM command extensions:
 *
 * The TPM spec doesn't really talk about vendor-defined ordinal
 * values, so the following makes some assumptions.  Here's how we
 * define ordinals for TPM extensions in the simulator: First, the
 * "Vendor" flag must be set in the ordinal (this is clear), and then
 * we'll treat the "purview" like a vendor code - to avoid possible
 * conflicts, we've just selected a random value for the purview,
 * shich should hopefully avoid collisions with other vendor-defined
 * ordinals.  Beyond that, we use the bottom 16 bits for extension
 * projects as follows: we assume that no extension project will need
 * more than 64 new command ordinals, so we use the bottom 6 bits for
 * the ordinal with a project, and the top 10 bits as the project
 * number.  Project numbers are assigned sequentially, and the latest
 * list of "official" extension numbers is given at
 * http://span.uncg.edu/tpmsim - if you are creating an extension that
 * you would like to share with others, so would like an official
 * extension number, contact Steve Tate at srtate@uncg.edu and a code
 * will be reserved for your project.
 */

#define TPM_SIMEXT_PURVIEW 0x6b

#define TPM_SIMEXT_PROJ_ORACLES 1
#define TPM_SIMEXT_PROJ_HASHTREE 2

#define TPM_SIMEXT_ORD_PROJ(p,o) (TPM_VENDOR_COMMAND|(TPM_SIMEXT_PURVIEW<<16)|(p<<6)|o)

BOOL key_algorithmID_is_genkey(TPM_ALGORITHM_ID id);
BOOL key_algorithmID_is_transkey(TPM_ALGORITHM_ID id);
BOOL key_es_is_validtrans(TPM_ENC_SCHEME es);

BOOL key_parms_is_supported(TPM_KEY_PARMS *parms);
BOOL key_parms_is_minimal_size(TPM_KEY_PARMS *parms);
BOOL key_parms_is_fips_size(TPM_KEY_PARMS *parms);
BOOL key_parms_is_strong_size(TPM_KEY_PARMS *parms);
BOOL key_parms_is_good_es(TPM_KEY_PARMS *parms);
BOOL key_parms_is_any_es(TPM_KEY_PARMS *parms);
BOOL key_parms_is_good_ss(TPM_KEY_PARMS *parms);
BOOL key_parms_is_idkey_ss(TPM_KEY_PARMS *parms);
BOOL key_parms_is_any_ss(TPM_KEY_PARMS *parms);
BOOL key_parms_requires_info(TPM_KEY_PARMS *parms);
BOOL key_key_is_valid_keylengths(TPM_KEY *key, TPM_STORE_ASYMKEY *store);

UINT32 key_parms_get_ctextsize(TPM_KEY_PARMS *parms);
UINT32 key_parms_get_sigsize(TPM_KEY_PARMS *parms);
UINT32 key_parms_get_ptextsizemax(TPM_KEY_PARMS *parms);

void key_parms_set_strong_keylen(TPM_KEY_PARMS *parms);
void key_parms_set_strong_es(TPM_KEY_PARMS *parms);

int key_algorithmID_init_parms(TPM_ALGORITHM_ID id, TPM_KEY_PARMS *parms);
int key_parms_init_parms(TPM_KEY_PARMS *dest, TPM_KEY_PARMS *src);

int key_parms_native_init_storepubkey(TPM_KEY_PARMS *parms, BYTE *native, 
				      TPM_STORE_PUBKEY *out);
int key_keydata_init_storepubkey(TPM_KEY_DATA *key, TPM_STORE_PUBKEY *out);

int key_pubkeydata_init_pubkey(TPM_PUBKEY_DATA *src, TPM_PUBKEY *dest);
int key_keydata_init_pubkey(TPM_KEY_DATA *key, TPM_PUBKEY *out);

int key_pubkey_init_pubkeydata(TPM_PUBKEY *src, TPM_PUBKEY_DATA *dest);
int key_keydata_init_pubkeydata(TPM_KEY_DATA *key, TPM_PUBKEY_DATA *out);

int key_parms_native_init_storeprivkey(TPM_KEY_PARMS *parms, BYTE *nativeKey,
				       TPM_STORE_PRIVKEY *privKey);

int key_keydata_init_keydata(TPM_KEY_DATA *in, TPM_KEY_DATA *out);

int key_parms_storekeys_init_native(TPM_KEY_PARMS *parms, 
				      TPM_STORE_PUBKEY *pubKey, 
				      TPM_STORE_PRIVKEY *privKey,
				      BYTE *native);

int key_parms_generate_native(TPM_KEY_PARMS *parms, BYTE *native);
int key_pubkeydata_encrypt(TPM_PUBKEY_DATA *key, BYTE *in, UINT32 in_size,
			   BYTE *enc, UINT32 *enc_size);
int key_keydata_decrypt(TPM_KEY_DATA *key, BYTE *enc, UINT32 enc_size,
			BYTE *out, UINT32 *out_size);
TPM_RESULT key_parms_native_sign(TPM_KEY_PARMS *parms, BYTE *nativeKey, 
			  const uint8_t *data, size_t data_len, uint8_t *sig);
TPM_RESULT key_keydata_sign(TPM_KEY_DATA *key, BYTE *inData, UINT32 inDataSize,
		     BOOL hashFirst, BYTE **sig, UINT32 *sigSize);
TPM_RESULT key_parms_nativepk_verify(TPM_KEY_PARMS *parms, BYTE *nativepk, 
		    BYTE *data, UINT32 dataSize, BOOL hashFirst, uint8_t *sig);

UINT32 key_parms_native_marshal_size(TPM_KEY_PARMS *parms, BYTE *native);
int key_parms_marshal_native(TPM_KEY_PARMS *parms, BYTE **ptr, UINT32 *length, 
			     BYTE *native);
int key_parms_unmarshal_native(TPM_KEY_PARMS *parms, BYTE **ptr, UINT32 *length,
			       BYTE *native);

UINT32 key_parms_nativepk_marshal_size(TPM_KEY_PARMS *parms, BYTE *nativepk);
int key_parms_marshal_nativepk(TPM_KEY_PARMS *parms, BYTE **ptr, 
			       UINT32 *length, BYTE *nativepk);
int key_parms_unmarshal_nativepk(TPM_KEY_PARMS *parms, BYTE **ptr, 
				 UINT32 *length, BYTE *nativepk);

int key_parms_marshal_algkeyparms(TPM_KEY_PARMS *parms, BYTE **ptr,
				  UINT32 *length);
int key_parms_unmarshal_algkeyparms(TPM_KEY_PARMS *parms, BYTE **ptr,
				    UINT32 *length);

void key_parms_free(TPM_KEY_PARMS *parms);
void key_parms_native_free(TPM_KEY_PARMS *parms, BYTE *bnativeKey);
void key_parms_nativepk_free(TPM_KEY_PARMS *parms, BYTE *bnativepk);

int key_selftest_all_keytypes(void);

/*
 * Function table for key type plugins - see documentation in keytype_null
 * for info on each function
 */

typedef struct keytype_info_s {
  TPM_ALGORITHM_ID algorithmID;
  const char *debugname;
  int (*init)(void);

  BOOL (*is_genkey)(void);
  BOOL (*is_transkey)(void);
  BOOL (*es_is_validtrans)(TPM_ENC_SCHEME es);

  BOOL (*parms_is_supported)(TPM_KEY_PARMS *parms);
  BOOL (*parms_is_minimal_size)(TPM_KEY_PARMS *parms);
  BOOL (*parms_is_fips_size)(TPM_KEY_PARMS *parms);
  BOOL (*parms_is_strong_size)(TPM_KEY_PARMS *parms);
  BOOL (*parms_is_good_es)(TPM_KEY_PARMS *parms);
  BOOL (*parms_is_any_es)(TPM_KEY_PARMS *parms);
  BOOL (*parms_is_good_ss)(TPM_KEY_PARMS *parms);
  BOOL (*parms_is_idkey_ss)(TPM_KEY_PARMS *parms);
  BOOL (*parms_is_any_ss)(TPM_KEY_PARMS *parms);
  BOOL (*parms_requires_info)(TPM_KEY_PARMS *parms);
  BOOL (*key_is_valid_keylengths)(TPM_KEY *key, TPM_STORE_ASYMKEY *store);

  UINT32 (*parms_get_ctextsize)(TPM_KEY_PARMS *parms);
  UINT32 (*parms_get_sigsize)(TPM_KEY_PARMS *parms);
  UINT32 (*parms_get_ptextsizemax)(TPM_KEY_PARMS *parms);

  void (*parms_set_strong_keylen)(TPM_KEY_PARMS *parms);
  void (*parms_set_strong_es)(TPM_KEY_PARMS *parms);

  int (*parms_set_default_algparms)(TPM_KEY_PARMS *parms);
  int (*parms_init_algparms)(TPM_KEY_PARMS *src, BYTE *destalgparms);

  int (*parms_native_init_storepubkey)(TPM_KEY_PARMS *parms, BYTE *native,
				       TPM_STORE_PUBKEY *out);
  int (*parms_nativepk_init_storepubkey)(TPM_KEY_PARMS *parms, BYTE *nativepk,
					 TPM_STORE_PUBKEY *out);

  int (*parms_native_init_storeprivkey)(TPM_KEY_PARMS *parms, BYTE *nativeKey,
					TPM_STORE_PRIVKEY *privKey);

  int (*parms_storekeys_init_native)(TPM_KEY_PARMS *parms, 
				       TPM_STORE_PUBKEY *spubkey,
				       TPM_STORE_PRIVKEY *sprivkey,
				       BYTE *native);
  int (*parms_native_init_native)(TPM_KEY_PARMS *parms, 
				    BYTE *nativein,
				    BYTE *nativeout);

  int (*pubkey_init_nativepk)(TPM_PUBKEY *src, BYTE *nativepk);
  int (*keydata_init_nativepk)(TPM_KEY_DATA *key, BYTE *nativepk);

  int (*parms_generate_native)(TPM_KEY_PARMS *parms, BYTE *native);
  int (*pubkeydata_encrypt)(TPM_PUBKEY_DATA *key, BYTE *in, UINT32 in_size,
			    BYTE *enc, UINT32 *enc_size);
  int (*keydata_decrypt)(TPM_KEY_DATA *key, BYTE *enc, UINT32 enc_size,
			 BYTE *out, UINT32 *out_size);
  TPM_RESULT (*parms_native_sign)(TPM_KEY_PARMS *parms, BYTE *nativeKey, 
			   const uint8_t *data, size_t data_len, 
			   BOOL hashFirst, uint8_t *sig);
  TPM_RESULT (*parms_nativepk_verify)(TPM_KEY_PARMS *parms, BYTE *nativepk, 
			       BYTE *data, UINT32 dataSize, 
			       BOOL hashFirst, uint8_t *sig);

  UINT32 (*parms_native_marshal_size)(TPM_KEY_PARMS *parms, BYTE *native);
  int (*parms_marshal_native)(TPM_KEY_PARMS *parms, BYTE **ptr, 
			      UINT32 *length, BYTE *native);
  int (*parms_unmarshal_native)(TPM_KEY_PARMS *parms, BYTE **ptr,
				UINT32 *length, BYTE *native);

  UINT32 (*parms_nativepk_marshal_size)(TPM_KEY_PARMS *parms, BYTE *nativepk);
  int (*parms_marshal_nativepk)(TPM_KEY_PARMS *parms, BYTE **ptr, 
				UINT32 *length, BYTE *nativepk);
  int (*parms_unmarshal_nativepk)(TPM_KEY_PARMS *parms, BYTE **ptr,
				  UINT32 *length, BYTE *nativepk);

  int (*parms_marshal_algkeyparms)(TPM_KEY_PARMS *parms, BYTE **ptr, 
				   UINT32 *length);
  int (*parms_unmarshal_algkeyparms)(TPM_KEY_PARMS *parms, BYTE **ptr, 
				     UINT32 *length);

  void (*parms_free)(TPM_KEY_PARMS *parms);
  void (*parms_native_free)(TPM_KEY_PARMS *parms, BYTE *native);
  void (*parms_nativepk_free)(TPM_KEY_PARMS *parms, BYTE *pnativepk);

  int (*selftest)(void);
} keytype_info;

extern keytype_info *key_typeInfo[];
extern int key_numKeyTypes;

void keytype_add(keytype_info *kt_info);

#endif
