/* Software-based Trusted Platform Module (TPM) Emulator
 * Emulator Copyright (C) 2004-2010 Mario Strasser <mast@gmx.net>
 * Timing/plugin code Copyright (C) 2010-2012 Stephen R. Tate <srtate@uncg.edu>
 *
 * This module is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 *
 * This module is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>

#include "plugins.h"

TPM_EXTENSION_DATA *plugin_permanent_data;
extern const char *tpm_storage_file;

static void plugin_native_to_blob(TPM_EXTENSION_DATA *data)
{
  if (data->fns == NULL)
    return;

  data->fns->native_init_data(data->native, &data->data, &data->size);
}


static int writeall(int fd, BYTE *data, size_t len)
{
  ssize_t res;
  while (len > 0) {
    res = write(fd, data, len);
    if (res < 0) {
      return -1;
    }
    len -= res; 
    data += res;
  }
  return 0;
}


int plugin_tpm_store_permanent_data(void)
{
  char fname[strlen(tpm_storage_file)+5];
  int fh;
  TPM_EXTENSION_DATA *curr_plugin_data = plugin_permanent_data;

  if (curr_plugin_data == NULL)
    return 0;

  strcpy(fname, tpm_storage_file);
  strcat(fname, "-ext");
#if defined(_WIN32) || defined(_WIN64)
  fh = open(fname, O_WRONLY | O_TRUNC | O_CREAT | O_BINARY, S_IRUSR | S_IWUSR); 
#else
  fh = open(fname, O_WRONLY | O_TRUNC | O_CREAT, S_IRUSR | S_IWUSR);
#endif
  if (fh < 0) return -1;
  while (curr_plugin_data != NULL) {
    if (curr_plugin_data->native_valid)
      plugin_native_to_blob(curr_plugin_data);

    if (curr_plugin_data->size > 0) {
      BYTE header[8], *hpos = header;
      UINT32 hlen = 8;
      tpm_marshal_UINT16(&hpos, &hlen, TPM_EXTENSION_TAG);
      tpm_marshal_UINT16(&hpos, &hlen, curr_plugin_data->ext_id);
      tpm_marshal_UINT32(&hpos, &hlen, curr_plugin_data->size);
      if (writeall(fh, header, sizeof(header))
	  || writeall(fh, curr_plugin_data->data, curr_plugin_data->size)) {
	close(fh);
	return -1;
      }
    }

    curr_plugin_data = curr_plugin_data->next;
  }
  close(fh);
  return 0;
}


int plugin_tpm_restore_permanent_data(void)
{
  char fname[strlen(tpm_storage_file)+5];
  int fh;
  ssize_t res;
  size_t total_length;
  uint8_t *data;
  size_t data_length;

  strcpy(fname, tpm_storage_file);
  strcat(fname, "-ext");

#if defined(_WIN32) || defined(_WIN64)
  fh = open(fname, O_RDONLY | O_BINARY); 
#else
  fh = open(fname, O_RDONLY);
#endif
  if (fh < 0) return -1;
  total_length = lseek(fh, 0, SEEK_END);
  lseek(fh, 0, SEEK_SET);
  data = tpm_malloc(total_length);
  if (data == NULL) {
    close(fh);
    return -1;
  }
  data_length = 0;
  while (total_length > 0) {
    res = read(fh, &data[data_length], total_length);
    if (res < 0) {
      close(fh);
      tpm_free(data);
      return -1;
    }
    if (res == 0) break;
    data_length += res;
    total_length -= res;
  }
  close(fh);

  total_length = data_length;
  BYTE *ptr = data;
  plugin_permanent_data = NULL;
  TPM_EXTENSION_DATA *last = NULL;
  while (total_length >= 8) {
    UINT16 thistag, thisid;
    UINT32 thislen;
    tpm_unmarshal_UINT16(&ptr, &total_length, &thistag);
    tpm_unmarshal_UINT16(&ptr, &total_length, &thisid);
    tpm_unmarshal_UINT32(&ptr, &total_length, &thislen);
    if (total_length < thislen) {
      tpm_free(data);
      return -1;
    }

    TPM_EXTENSION_DATA *thisext = tpm_malloc(sizeof(TPM_EXTENSION_DATA));
    thisext->ext_id = thisid;
    thisext->fns = NULL;
    thisext->size = thislen;
    thisext->data = tpm_malloc(thislen);
    thisext->native_valid = 0;
    thisext->native = NULL;
    thisext->next = NULL;

    tpm_unmarshal_BYTE_ARRAY(&ptr, &total_length, thisext->data, thislen);

    if (last == NULL)
      plugin_permanent_data = thisext;
    else
      last->next = thisext;
    last = thisext;
  }


  return 0;
}


BYTE *plugin_get_permdata(UINT16 ext_id, TPM_EXTENSION_FUNCTIONS *functions)
{
  TPM_EXTENSION_DATA *curr = plugin_permanent_data;
  TPM_EXTENSION_DATA *last = NULL;

  while ((curr != NULL) && (curr->ext_id != ext_id)) {
    last = curr;
    curr = curr->next;
  }

  if ((curr == NULL) && (functions != NULL)) {
    curr = tpm_malloc(sizeof(TPM_EXTENSION_DATA));
    curr->ext_id = ext_id;
    curr->fns = functions;
    curr->size = 0;
    curr->data = NULL;
    curr->native_valid = 1;
    curr->native = functions->init_native();
    curr->next = NULL;

    if (last == NULL)
      plugin_permanent_data = curr;
    else
      last->next = curr;
  }

  if ((curr != NULL) && (curr->fns == NULL)) {
    curr->fns = functions;
  }

  if ((curr != NULL) && (!curr->native_valid)) {
    curr->native = curr->fns->data_init_native(curr->data, curr->size);
    if (curr->native != NULL)
      curr->native_valid = 1;
  }

  if (curr != NULL)
    return curr->native;
  else
    return NULL;
}
