/* Software-based Trusted Platform Module (TPM) Emulator
 * Copyright (C) 2004-2010 Mario Strasser <mast@gmx.net>
 * Timing/plugin code Copyright (C) 2010-2012 Stephen R. Tate <srtate@uncg.edu>
 *
 * This module is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 *
 * This module is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 */

#include "tpm_emulator.h"
#include "tpm_commands.h"
#include "tpm_marshalling.h"
#include "tpm_data.h"
#include "crypto/sha1.h"
#include "crypto/rsa.h"

/*
 * Maintenance Functions ([TPM_Part3], Section 12)
 */

static int tpm_setup_privkey(TPM_KEY_DATA *key, TPM_KEY *privkey)
{
  privkey->tag = TPM_TAG_KEY12;
  privkey->fill = 0;
  privkey->keyUsage = key->keyUsage;
  privkey->keyFlags = key->keyFlags;
  privkey->authDataUsage = key->authDataUsage;
  if (key_parms_init_parms(&key->algorithmParms, &privkey->algorithmParms))
      return -1;
  memcpy(&privkey->PCRInfo, &key->pcrInfo, sizeof(TPM_PCR_INFO)); 
  privkey->PCRInfoSize = sizeof_TPM_PCR_INFO(privkey->PCRInfo);
  privkey->encDataSize = 0;
  privkey->encData = NULL;
  if (key_keydata_init_storepubkey(key, &privkey->pubKey)) {
      free_TPM_KEY_PARMS(privkey->algorithmParms);
      return -1;
  }
  return 0;
}

TPM_RESULT TPM_CreateMaintenanceArchive(BOOL generateRandom, TPM_AUTH *auth1,
                                        UINT32 *randomSize, BYTE **random,
                                        UINT32 *archiveSize, BYTE **archive)
{
  TPM_RESULT res;
  TPM_KEY key;
  TPM_DIGEST key_digest;
  TPM_STORE_PRIVKEY store;
  BYTE *buf, *ptr;
  UINT32 len;
  size_t buf_len;
  
  info("TPM_CreateMaintenanceArchive()");
  if (!tpmData.permanent.flags.allowMaintenance) return TPM_DISABLED_CMD;
  res = tpm_verify_auth(auth1, tpmData.permanent.data.ownerAuth, TPM_KH_OWNER);
  if (res != TPM_SUCCESS) return res;
  if (!tpmData.permanent.data.manuMaintPub.valid) return TPM_KEYNOTFOUND;
  /* set up a TPM_KEY structure for the SRK */
  if (tpm_setup_privkey(&tpmData.permanent.data.srk, &key) != 0) {
    debug("tpm_setup_privkey(SRK) failed");
    return TPM_FAIL;
  }
  if (tpm_compute_key_digest(&key, &key_digest) != 0) {
    debug("tpm_compute_key_digest() failed");
    free_TPM_KEY(key);
    return TPM_FAIL;
  }
  /* generate an OAEP encoding of the TPM_MIGRATE_ASYMKEY structure for
     the SRK: 0x00|seed|0x00-pad|0x01|TPM_MIGRATE_ASYMKEY */
  /* Note: comment above is from original code - it's not correct though */
  debug("generating OAEP encoding");
  if (key_parms_native_init_storeprivkey(&tpmData.permanent.data.srk.algorithmParms,
					 tpmData.permanent.data.srk.bnativekey, 
					 &store)) {
      debug("initializing key structure failed.");
      free_TPM_KEY(key);
      return TPM_FAIL;
  }
  if (store.keyLength <= 16) {
      debug("stored private key length is <= 16");
      free_TPM_KEY(key);
      return TPM_FAIL;
  }
  len = buf_len = 45+(store.keyLength-16)+41;
  ptr = buf = tpm_malloc(buf_len);
  if (buf == NULL) {
    free_TPM_STORE_PRIVKEY(store);
    free_TPM_KEY(key);
    return TPM_NOSPACE;
  }
  memset(buf, 0, buf_len);
  tpm_marshal_UINT32(&ptr, &len, store.keyLength);
  memcpy(ptr, store.key, 16);
  ptr += 16;
  memcpy(ptr, tpmData.permanent.data.ownerAuth, sizeof(TPM_SECRET));
  len = 46 + store.keyLength - 16;
  ptr = &buf[buf_len - len];
  tpm_marshal_BYTE(&ptr, &len, 0x01);  
  tpm_marshal_TPM_PAYLOAD_TYPE(&ptr, &len, TPM_PT_MAINT);
  tpm_marshal_TPM_NONCE(&ptr, &len, &tpmData.permanent.data.tpmProof);
  tpm_marshal_TPM_DIGEST(&ptr, &len, &key_digest);
  tpm_marshal_UINT32(&ptr, &len, store.keyLength - 16); 
  memcpy(ptr, &store.key[16], store.keyLength-16);
  tpm_rsa_mask_generation(buf, SHA1_DIGEST_LENGTH, 
			  &buf[SHA1_DIGEST_LENGTH], buf_len - SHA1_DIGEST_LENGTH);
  tpm_rsa_mask_generation(&buf[SHA1_DIGEST_LENGTH], 
			  buf_len - SHA1_DIGEST_LENGTH, buf, SHA1_DIGEST_LENGTH);

  /* XOR encrypt OAEP encoding */
  debug("generateRandom = %d", generateRandom);
  if (generateRandom) {
    *randomSize = buf_len;
    *random = tpm_malloc(*randomSize);
    if (*random == NULL) {
      free_TPM_STORE_PRIVKEY(store);
      free_TPM_KEY(key);
      tpm_free(buf);
      return TPM_NOSPACE;
    }
    tpm_get_random_bytes(*random, *randomSize);
    for (len = 0; len < buf_len; len++) buf[len] ^= (*random)[len];
  } else {
    *randomSize = 0;
    *random = NULL;
    tpm_rsa_mask_generation(tpmData.permanent.data.ownerAuth,
                            SHA1_DIGEST_LENGTH, buf, buf_len);
  }
  /* Encrypt OAEP encoding */
  if (key_pubkeydata_encrypt(&tpmData.permanent.data.manuMaintPub,
			     buf, buf_len, buf, &buf_len) != 0) {
    debug("key_pubkeydata_encrypt() failed");
    free_TPM_STORE_PRIVKEY(store);
    free_TPM_KEY(key);
    tpm_free(buf);
    return TPM_FAIL;
  }
  key.encData = buf;
  key.encDataSize = buf_len;
  /* marshal response */
  len = *archiveSize = sizeof_TPM_KEY(key);
  ptr = *archive = tpm_malloc(len);
  debug("archiveSize = %d, archive = %p", *archiveSize, *archive);
  if (ptr == NULL || tpm_marshal_TPM_KEY(&ptr, &len, &key)) {
    tpm_free(ptr);
    tpm_free(*random);
    free_TPM_STORE_PRIVKEY(store);
    free_TPM_KEY(key);
    return TPM_NOSPACE;
  }
  free_TPM_STORE_PRIVKEY(store);
  free_TPM_KEY(key);
  *archiveSize -= len;
  tpmData.permanent.flags.maintenanceDone = TRUE;
  return TPM_SUCCESS;
}

TPM_RESULT TPM_LoadMaintenanceArchive(UINT32 archiveSize, BYTE *archive,
                                      UINT32 sigSize, BYTE *sig,
                                      UINT32 randomSize, BYTE *random,
                                      TPM_AUTH *auth1)
{
  TPM_RESULT res;
  TPM_KEY newsrk;
  TPM_DIGEST digest;
  TPM_KEY_DATA *srk;
  TPM_STORE_PRIVKEY store;
  tpm_sha1_ctx_t sha1;
  BYTE *buf, *ptr;
  UINT32 len;
  size_t buf_len;
  UINT32 keylen;

  info("TPM_LoadMaintenanceArchive()");
  /* verify authorization */
  if (!tpmData.permanent.data.manuMaintPub.valid) return TPM_KEYNOTFOUND;
  if (!tpmData.permanent.flags.allowMaintenance) return TPM_DISABLED_CMD;
  res = tpm_verify_auth(auth1, tpmData.permanent.data.ownerAuth, TPM_KH_OWNER);
  if (res != TPM_SUCCESS) return res;
  /* verify signature */
  tpm_sha1_init(&sha1);
  tpm_sha1_update(&sha1, archive, archiveSize);
  tpm_sha1_final(&sha1, digest.digest);
  if (sigSize != key_parms_get_sigsize(&tpmData.permanent.data.manuMaintPub.algorithmParms))
    return TPM_BAD_SIGNATURE;

  res = key_parms_nativepk_verify(&tpmData.permanent.data.manuMaintPub.algorithmParms,
			  tpmData.permanent.data.manuMaintPub.bnativepubkey,
			  digest.digest, sizeof(digest.digest), TRUE, sig);
  if (res != TPM_SUCCESS) return res;
  /* unmarshal archive */
  ptr = archive; len = archiveSize;
  if (tpm_unmarshal_TPM_KEY(&ptr, &len, &newsrk)
      || len != 0)
    return TPM_BAD_PARAMETER;

  /* validate new SRK */
  if (newsrk.keyFlags & TPM_KEY_FLAG_MIGRATABLE
      || newsrk.keyUsage != TPM_KEY_STORAGE)
    return TPM_INVALID_KEYUSAGE;
  if (!key_parms_is_supported(&(newsrk.algorithmParms))
      || !key_parms_is_strong_size(&(newsrk.algorithmParms))
      || !key_parms_is_good_es(&(newsrk.algorithmParms))
      || newsrk.PCRInfoSize != 0)
    return TPM_BAD_KEY_PROPERTY;

  /* decrypt private key */
  buf_len = newsrk.encDataSize;
  buf = tpm_malloc(buf_len);
  if (buf == NULL) return TPM_NOSPACE;
  srk = &tpmData.permanent.data.srk;
  if (key_keydata_decrypt(srk, newsrk.encData, newsrk.encDataSize, 
			  buf, &buf_len)
      || ((randomSize > 0) && (randomSize != buf_len))) {
    debug("key_keydata_decrypt() failed");
    tpm_free(buf);
    return TPM_DECRYPT_ERROR;
  }
  if (randomSize > 0) {
    for (len = 0; len < buf_len; len++) buf[len] ^= random[len];
  } else {
    tpm_rsa_mask_generation(tpmData.permanent.data.ownerAuth,
                            SHA1_DIGEST_LENGTH, buf, buf_len);
  }
  tpm_rsa_mask_generation(&buf[SHA1_DIGEST_LENGTH], buf_len-SHA1_DIGEST_LENGTH,
			  buf, SHA1_DIGEST_LENGTH);
  tpm_rsa_mask_generation(buf, SHA1_DIGEST_LENGTH, &buf[SHA1_DIGEST_LENGTH], 
			  buf_len - SHA1_DIGEST_LENGTH);

  /* Check mid buffer for consistency with OAEP encoding of maint archive */
  /* Must look ahead so we don't clobber owner auth for invalid archive */
  for (ptr = &buf[20 + sizeof(TPM_SECRET)]; *ptr == 0x00; ptr++);
  if (ptr[0] != 0x01 || ptr[1] != TPM_PT_MAINT) {
    debug("OAEP encoding is invalid");
    tpm_free(buf);
    return TPM_DECRYPT_ERROR;
  }

  ptr = buf; len = buf_len;
  tpm_unmarshal_UINT32(&ptr, &len, &keylen);
  debug("actual buffer len = %d ; calculated = %d", buf_len, 70+keylen);
  ptr += 16;  len -= 16;   // Just skip of K1 data for now

  /* clear owner and set new ownerAuth */
  tpm_owner_clear();
  tpm_unmarshal_TPM_SECRET(&ptr, &len, &tpmData.permanent.data.ownerAuth);

  for (ptr = &buf[20 + sizeof(TPM_SECRET)]; *ptr == 0x00; ptr++, len--);
  ptr += 2;  len -= 2;

  /* update tpmProof */
  tpm_unmarshal_TPM_NONCE(&ptr, &len, &tpmData.permanent.data.tpmProof);

  /*
   * TODO: Need to check pubkeydigest -- skipping over for now until more
   *       testing is possible
   */

  ptr += 20;   len -= 20;

  tpm_unmarshal_UINT32(&ptr, &len, &store.keyLength);
  if (store.keyLength+16 != keylen) {
    debug("TPM_LoadMaintenanceArchive(): key length mismatch");
    tpm_free(buf);
    return TPM_FAIL;
  }
  store.key = tpm_malloc(keylen);
  if (store.key == NULL) {
    tpm_free(buf);
    return TPM_NOSPACE;
  }
  memcpy(store.key, &buf[4], 16);
  memcpy(&store.key[16], ptr, store.keyLength);
  store.keyLength = keylen;

  /* update SRK */
  /* memset is redundant - should have been done by tpm_owner_clear()... */
  memset(srk, 0, sizeof(TPM_KEY_DATA));
  srk->keyUsage = newsrk.keyUsage;
  srk->keyFlags = newsrk.keyFlags;
  srk->keyFlags |= TPM_KEY_FLAG_PCR_IGNORE;
  srk->keyFlags &= ~TPM_KEY_FLAG_HAS_PCR;
  srk->keyControl = TPM_KEY_CONTROL_OWNER_EVICT;
  srk->authDataUsage = newsrk.authDataUsage;

  memcpy(srk->usageAuth, tpmData.permanent.data.ownerAuth, sizeof(TPM_SECRET));
  memcpy(srk->migrationAuth, tpmData.permanent.data.tpmProof.nonce,
	 sizeof(TPM_SECRET));

  if (key_parms_init_parms(&newsrk.algorithmParms, &srk->algorithmParms)) {
    debug("Setting up parms for new srk failed");
    tpm_free(store.key);
    tpm_free(buf);
    return TPM_FAIL;
  }

  srk->parentPCRStatus = FALSE;

  if (key_parms_storekeys_init_native(&srk->algorithmParms, &newsrk.pubKey, 
				      &store, srk->bnativekey)) {
    tpm_free(store.key);
    free_TPM_KEY_PARMS(srk->algorithmParms);
    tpm_free(buf);
    debug("setting new srk native key failed");
    return TPM_FAIL;
  }

  /* enable SRK and mark TPM as owned */
  srk->payload = TPM_PT_ASYM;
  tpmData.permanent.flags.owned = TRUE;
  tpmData.permanent.flags.maintenanceDone = TRUE;
  tpm_free(store.key);
  tpm_free(buf);
  return TPM_SUCCESS;
}

TPM_RESULT TPM_KillMaintenanceFeature(TPM_AUTH *auth1)
{
  TPM_RESULT res;

  info("TPM_KillMaintenanceFeature()");
  res = tpm_verify_auth(auth1, tpmData.permanent.data.ownerAuth, TPM_KH_OWNER);
  if (res != TPM_SUCCESS) return res;
  tpmData.permanent.flags.allowMaintenance = FALSE;
  return TPM_SUCCESS;
}

TPM_RESULT TPM_LoadManuMaintPub(TPM_NONCE *antiReplay, TPM_PUBKEY *pubKey,
                                TPM_DIGEST *checksum)
{
  TPM_PUBKEY_DATA *key = &tpmData.permanent.data.manuMaintPub;

  info("TPM_LoadManuMaintPub()");
  if (key->valid) return TPM_DISABLED_CMD;
  if (!key_parms_is_supported(&(pubKey->algorithmParms))
      || !key_parms_is_good_es(&(pubKey->algorithmParms)))
    return TPM_BAD_KEY_PROPERTY;
  if (key_pubkey_init_pubkeydata(pubKey, key))
      return TPM_FAIL;
  if (tpm_compute_pubkey_checksum(antiReplay, pubKey, checksum) != 0) {
      free_TPM_PUBKEY_DATA((*key));
      key->valid = 0;
      return TPM_FAIL;
  }
  return TPM_SUCCESS;
}


TPM_RESULT TPM_ReadManuMaintPub(TPM_NONCE *antiReplay, TPM_DIGEST *checksum)
{
  TPM_PUBKEY key;
  int res;
  
  info("TPM_ReadManuMaintPub()");
  if (!tpmData.permanent.data.manuMaintPub.valid) return TPM_KEYNOTFOUND;
  if (key_pubkeydata_init_pubkey(&tpmData.permanent.data.manuMaintPub, &key))
    return TPM_FAIL;
  res = tpm_compute_pubkey_checksum(antiReplay, &key, checksum);
  free_TPM_PUBKEY(key);
  return (res == 0) ? TPM_SUCCESS : TPM_FAIL;
}

